import datetime
import resend
import time

resend.api_key = "re_X1PBTBvD_5mJfFM98AuF2278fNAGfXVNV"
from termcolor import colored


class Alerting:

  def __init__(self,
               alert_threshold=5,
               send_notifications=True,
               user_emails=set()):
    self.unhandled_errors = {
    }  # dictionary of openai errors + frequency count for that hour
    now = datetime.datetime.now()
    self.current_time_block = now.hour
    self.set_cooldown = False
    self.cooldown_start_time = time.time()
    self.send_notifications = send_notifications
    self.user_emails = user_emails

  def start_cooldown(self):
    self.set_cooldown = True
    self.cooldown_start_time = time.time()

  def send_alert(self, error_type, most_recent_error):
    print(colored("Sending alert", "magenta"))
    params = {
      "from":
      "krrish@berri.ai",
      "subject":
      """reliableGPT 💪: Unhandled Error - {}""".format(error_type),
      "html":
      """
            <p><strong>ReliableGPT Error:</strong></p>
            <p>A critical error occurred that ReliableGPT was unable to handle.</p>
            <p>This error has happened 5 times in the last hour:</p>
            <p><strong>{}</strong></p>
            <p> This was your most recent error {} </p>
            <p>Suggest a way we could cover this error with reliableGPT 💪:
        <a href="https://github.com/BerriAI/reliableGPT/issues/new">Here</a></p>

        <p>Join our Discord for Support:
        <a href=" https://discord.com/invite/WXFfTeEXRh">Here</a></p>
            """.format(error_type, most_recent_error)
    }
    for email in self.user_emails:
      params["to"] = email
      email = resend.Emails.send(params)
    return

  def add_emails(self, user_email):
    if type(user_email) == list:
      for email in user_email:
        self.user_emails.add(email)
    else:
      self.user_emails.add(user_email)
    return

  def add_error(self, openai_error):
    if openai_error != None:
      openai_error = openai_error.error  # index into the error attribute of the class
    error_type = None  # defalt to being None
    if openai_error != None and 'type' in openai_error:
      error_type = openai_error['type']
    now = datetime.datetime.now()
    curr_time = now.hour

    if curr_time == self.current_time_block:
      if error_type in self.unhandled_errors:
        self.unhandled_errors[error_type] += 1
        print(f"Got unhandled error {self.unhandled_errors}")
      else:
        self.unhandled_errors[error_type] = 1
      if self.unhandled_errors[error_type] >= 5:
        self.unhandled_errors[error_type] = 0  # reset this to 0
        # cooldown for 15 minutes -> do this to prevent email spam.
        if self.set_cooldown and time.time() - self.cooldown_start_time > 900:
          self.send_alert(error_type, openai_error)
          self.set_cooldown()
        elif self.set_cooldown == False:
          self.send_alert(error_type, openai_error)
          self.start_cooldown()
    else:  # reset
      self.unhandled_errors[error_type] = 1
      self.current_time_block = curr_time
    return



# reliablegpt/__init__.py
# from .RateLimitHandler_helpers import *
# from .Alerting import *
# from .IndividualRequest import *
# from .Model import *
# from .RateLimitHandler import *
from .main import *  # Import all the symbols from main.py





import queue
from concurrent.futures import ThreadPoolExecutor
import threading
from termcolor import colored
import traceback


from reliablegpt.CustomQueue import CustomQueue


class APICallHandler():

  def __init__(self, max_token_capacity, max_request_capacity, verbose):
    super().__init__()
    self.verbose = verbose
    self.q = CustomQueue(max_token_capacity=max_token_capacity,
                         max_request_capacity=max_request_capacity,
                         verbose=verbose)
    self.results = {
    }  #key-value pair, task_id is the key, and the response from openai is the value
    # Define a sentinel value for stopping the worker
    # Start the worker thread
    worker_thread = threading.Thread(target=self.worker)
    self.worker_thread = None
    self.worker_running = False
    self.error_occurred = False
    self.error_description = None
    self.start_worker()

  def start_worker(self):
    if not self.worker_running:
      self.worker_thread = threading.Thread(target=self.worker)
      self.worker_thread.start()
      self.worker_running = True

  def stop_worker(self):
    if self.worker_running:
      self.worker_running = False
    return

  def print_verbose(self, print_statement):
    if self.verbose:
      print(colored("APICallHandler: " + str(print_statement), "cyan"))

  def result_handler(self, result, task_id):
    self.results[task_id] = result

  def process_item(self, function, kwargs, task_id):
    # wrap this around reliableGPT
    try:
      self.print_verbose(f"processing function: {function}")
      result = function(**kwargs)
      self.result_handler(result, task_id)
      return result
    except Exception as e:
      self.print_verbose(f"Error in task {task_id}: {e}")
      self.error_occurred = True
      self.error_description = traceback.format_exc()

  def run_async(self, items):
    self.print_verbose(items)
    if isinstance(items, list):
      threads = []
      for item in items:
        self.print_verbose(item)
        function, argument, task_id = item
        thread = threading.Thread(target=self.process_item,
                                  args=(function, argument, task_id))
        thread.start()
        threads.append(thread)

      # Wait for all threads to finish
      for thread in threads:
        thread.join

      # Check if any errors occurred
      if self.error_occurred:
        # Break out of the worker threads
        for thread in threads:
          thread.join()
        self.print_verbose("Errors occurred. Terminating worker threads.")

  # Define worker function that processes items from the queue
  def worker(self):
    timeout = 5
    try:
      self.error_occurred = False  # reset to False -> if one request is bad, don't stop any new request from coming in, in production. Just drop that batch. [TODO: drop just that single request].
      while True and self.error_occurred == False:
        try:
          items = self.q.get(timeout=timeout)  # Add timeout to the get method
        except queue.Empty:  # Add an exception handler for queue empty

          self.stop_worker()

          break
        except Exception as e:
          self.print_verbose(f"Hit an exception {traceback.format_exc()}")
        self.print_verbose(f"gets item from queue: {items}")
        if isinstance(items, list):
          self.run_async(items=items)
        else:
          function, argument, task_id = items
          self.process_item(function, argument, task_id)
        self.q.task_done()
    except:
      self.print_verbose(f"Hit an exception {traceback.format_exc()}")

  def add_sentinel(self, sentinel_value):
    self.q.put(sentinel_value)

  def add_task(self, process_func, input, task_id):
    if not self.worker_running:
      self.start_worker()
    self.q.put((process_func, input, task_id))

  def get_result(self, task_id):
    if task_id in self.results:
      return self.results[task_id]
    elif self.error_occurred:
      raise RuntimeError(
        f"An error occurred while processing one of the requests. Set verbose=True to debug this further. \n\n {self.error_description}"
      )
    else:
      return None

  def get_results(self):
    return self.results



from termcolor import colored
import requests
import copy
import openai
import posthog


class IndividualRequest:
  """A brief description of the class."""

  def __init__(self,
               model=None,
               fallback_strategy=[
                 'gpt-3.5-turbo', 'text-davinci-003', 'gpt-4',
                 'text-davinci-002'
               ],
               graceful_string="Sorry, the OpenAI API is currently down",
               user_email="",
               user_token="",
               send_notification=False,
               logging_fn=None,
               verbose=False):
    # Initialize instance variables
    self.model = model
    self.model_function = model.get_model_function()
    self.verbose = verbose
    self.graceful_string = graceful_string
    self.fallback_strategy = fallback_strategy
    self.user_email = user_email
    self.user_token = user_token
    self.save_request = logging_fn

    self.print_verbose(f"INIT fallback strategy {self.fallback_strategy}")

  def print_verbose(self, print_statement):
    if self.verbose:
      print(colored("Individual Request: " + print_statement, "blue"))

  ## Code that handles / wraps openai calls
  def __call__(self, *args, **kwargs):
    try:
      self.print_verbose(f"calling model function: {self.model_function}")
      self.print_verbose(f"these are the kwargs: {kwargs}")
      result = self.model_function(*args, **kwargs)
      self.print_verbose(f"This is the result: {str(result)[:500]}")
      return result
    except Exception as e:
      self.print_verbose("catches the error")
      return self.handle_exception(args, kwargs, e)

  def fallback_request(self, args, kwargs, fallback_strategy):
    result = None
    for model in fallback_strategy:
      new_kwargs = copy.deepcopy(kwargs)  # Create a deep copy of kwargs
      new_kwargs['model'] = model  # Update the model
      result = self.make_LLM_request(new_kwargs)
      if result != None:
        return result
    return None

  def make_LLM_request(self, new_kwargs):
    embedding_model = self.model.get_original_embeddings()
    chat_model = self.model.get_original_chat()
    completion_model = self.model.get_original_completion()
    try:
      if "embedding" in str(self.model_function):
        # retry embedding with diff key
        print(colored(f"ReliableGPT: Retrying Embedding request", "blue"))
        return embedding_model(**new_kwargs)

      model = str(new_kwargs['model'])
      print(
        colored(f"ReliableGPT: Checking request model {model} {new_kwargs}",
                "blue"))
      if "3.5" in model or "4" in model:  # call ChatCompletion
        print(
          colored(
            f"ReliableGPT: Retrying request with model CHAT {model} {new_kwargs}",
            "blue"))
        return chat_model(**new_kwargs)
      else:
        print(
          colored(f"ReliableGPT: Retrying request with model TEXT {model}",
                  "blue"))
        new_kwargs['prompt'] = " ".join(
          [message["content"] for message in new_kwargs['messages']])
        new_kwargs.pop('messages',
                       None)  # remove messages for completion models
        return completion_model(**new_kwargs)
    except Exception as e:
      print(colored(f"ReliableGPT: Got 2nd AGAIN Error {e}", "red"))
      raise ValueError(e)

  def api_key_handler(self, args, kwargs, fallback_strategy, user_email,
                      user_token):
    try:
      url = f"https://reliable-gpt-backend-9gus.zeet-berri.zeet.app/get_keys?user_email={user_email}&user_token={user_token}"
      response = requests.get(url)
      if response.status_code == 200:
        result = response.json()
        if result['status'] == 'failed':
          print(
            colored(
              f"ReliableGPT: No keys found for user: {user_email}, token: {user_token}",
              "red"))
          return None

        fallback_keys = result['response'][
          'openai_api_keys']  # list of fallback keys
        if len(fallback_keys) == 0:
          return None
        for fallback_key in fallback_keys:
          openai.api_key = fallback_key
          result = self.make_LLM_request(kwargs)
          if result != None:
            return result
      else:
        print(
          colored(
            f"ReliableGPT: No keys found for user: {user_email}, token: {user_token}",
            "red"))
      return None
    except Exception as e:
      raise ValueError(e)

  def handle_openAI_error(self,
                          args,
                          kwargs,
                          openAI_error,
                          fallback_strategy,
                          graceful_string,
                          user_email="",
                          user_token=""):
    # Error Types from https://platform.openai.com/docs/guides/error-codes/python-library-error-types
    # 1. APIError - retry, retry with fallback
    # 2. Timeout - retry, retry with fallback
    # 3. RateLimitError - retry, retry with fallback
    # 4. APIConnectionError - Check your network settings, proxy configuration, SSL certificates, or firewall rules.
    # 5. InvalidRequestError - User input was bad: context_length_exceeded,
    # 6. AuthenticationError - API key not working, return default hardcoded message
    # 7. ServiceUnavailableError - retry, retry with fallback
    self.print_verbose(
      colored(f"Inside handle openai error for User Email: {user_email}",
              "red"))
    if openAI_error != None:
      openAI_error = openAI_error.error  # index into the error attribute of the class

    error_type = None  # defalt to being None
    if openAI_error != None and 'type' in openAI_error:
      error_type = openAI_error['type']
    if error_type == 'invalid_request_error' or error_type == 'InvalidRequestError':
      # check if this is context window related, try with a 16k model
      if openAI_error.code == 'context_length_exceeded':
        print(
          colored(
            "ReliableGPT: invalid request error - context_length_exceeded",
            "red"))
        fallback_strategy = ['gpt-3.5-turbo-16k'] + fallback_strategy
        result = self.fallback_request(args=args,
                                       kwargs=kwargs,
                                       fallback_strategy=fallback_strategy)
        if result == None:
          return graceful_string
        else:
          return result
      if openAI_error.code == "invalid_api_key":
        print(
          colored("ReliableGPT: invalid request error - invalid_api_key",
                  "red"))
        result = self.api_key_handler(args=args,
                                      kwargs=kwargs,
                                      fallback_strategy=fallback_strategy,
                                      user_email=user_email,
                                      user_token=user_token)
        if result == None:
          return graceful_string
        else:
          return result

    # todo: alert on user_email that there is now an auth error
    elif error_type == 'authentication_error' or error_type == 'AuthenticationError':
      print(colored("ReliableGPT: Auth error", "red"))
      return graceful_string

    # catch all
    result = self.fallback_request(args=args,
                                   kwargs=kwargs,
                                   fallback_strategy=fallback_strategy)
    if result == None:
      return graceful_string
    else:
      return result
    return graceful_string

  def handle_exception(self, args, kwargs, e):
    result = self.graceful_string  # default to graceful string
    try:
      # Attempt No. 1, exception is received from OpenAI
      print(colored(f"ReliableGPT: Got Exception {e}", 'red'))
      result = self.handle_openAI_error(
        args=args,
        kwargs=kwargs,
        openAI_error=e,
        fallback_strategy=self.fallback_strategy,
        graceful_string=self.graceful_string,
        user_email=self.user_email,
        user_token=self.user_token)
      print(
        colored(f"ReliableGPT: Recovered got a successful response {result}",
                "green"))
      if result == self.graceful_string:
        # did a retry but still returned graceful string
        print("returns graceful string")
        self.save_request(
          user_email=self.user_email,
          graceful_string=self.graceful_string,
          posthog_event='reliableGPT.recovered_request_exception',
          result=result,
          posthog_metadata={
            'error': str(e),
            'recovered_response': result
          },
          errors=[e])
      else:
        # No errors, successfull retry
        self.save_request(user_email=self.user_email,
                          graceful_string=self.graceful_string,
                          posthog_event="reliableGPT.recovered_request",
                          result=result,
                          posthog_metadata={
                            'error': str(e),
                            'recovered_response': result
                          },
                          errors=[e])
    except Exception as e2:
      # Exception 2, After trying to rescue
      print("gets 2nd error: ", e2)
      self.save_request(
        user_email=self.user_email,
        graceful_string=self.graceful_string,
        posthog_event='reliableGPT.recovered_request_exception',
        result="",
        posthog_metadata={
          'original_error': str(e),
          'error2': str(e2),
          'recovered_response': self.graceful_string
        },
        errors=[e, e2])
      return result
    return result



# Prod Imports
from reliablegpt.IndividualRequest import IndividualRequest
from reliablegpt.RateLimitHandler import RateLimitHandler
from reliablegpt.Model import Model
from reliablegpt.alerting import Alerting

# # Dev Imports
# from IndividualRequest import IndividualRequest
# from RateLimitHandler import RateLimitHandler
# from Model import Model
# from alerting import Alerting

from posthog import Posthog

posthog = Posthog(
  project_api_key='phc_yZ30KsPzRXd3nYaET3VFDmquFKtMZwMTuFKVOei6viB',
  host='https://app.posthog.com')

alerting = Alerting()

# Parent Logging function
def save_request(user_email,
                 graceful_string,
                 posthog_event="",
                 result="",
                 posthog_metadata={},
                 errors=[]):
  try:
    #metadata  = kwargs['metadata']
    if posthog_event != "":
      posthog.capture(user_email, posthog_event,
                      posthog_metadata)  # save posthog event
    if result == graceful_string or len(
        errors) == 2:  # returns a graceful string or got a 2nd exception
      # send an email and alert
      for error in errors:
        alerting.add_error(error)
      # send_emails_task(self.user_email, posthog_metadata, self.send_notification)
  except:
    return  # safe function, should not impact error handling if logging fails


def reliableGPT(openai_create_function,
           fallback_strategy=[
             'gpt-3.5-turbo', 'text-davinci-003', 'gpt-4', 'text-davinci-002'
           ],
           graceful_string="Sorry, the OpenAI API is currently down",
           user_email="",
           user_token="",
           send_notification=True,
           model_limits_dir={},
           queue_requests=False):
  """Determine if an instantiation is calling the rate limit handler or the individual request wrapper directly and return the correct object"""

  primary_email = ""  # which api key management is mapped to
  ## Add email for alerting
  if isinstance(user_email, str):
    if user_email == "":
      raise ValueError("ReliableGPT Error: Please pass in a user email")
    else:
      alerting.add_emails(user_email)
      primary_email = user_email
  elif isinstance(user_email, list):
    primary_email = user_email[0]
    for email in user_email:
      alerting.add_emails(email)

  model = Model(openai_create_function)
  ## Route to the right object
  if queue_requests == False:
    return IndividualRequest(model,
                             fallback_strategy=fallback_strategy,
                             graceful_string=graceful_string,
                             user_email=primary_email,
                             user_token=user_token,
                             logging_fn=save_request,
                             send_notification=send_notification)
  elif queue_requests == True:
    max_token_capacity = 40000
    max_request_capacity = 200
    ## [TODO]: Allow handling multiple model types (gpt-3.5-turbo AND gpt-4)
    if model_limits_dir and isinstance(model_limits_dir, dict):
      keys = list(model_limits_dir.keys())
      max_token_capacity = model_limits_dir[keys[0]]["max_token_capacity"]
      max_request_capacity = model_limits_dir[keys[0]]["max_request_capacity"]

    return RateLimitHandler(model,
                            fallback_strategy=fallback_strategy,
                            graceful_string=graceful_string,
                            user_email=primary_email,
                            user_token=user_token,
                            send_notification=send_notification,
                            max_request_capacity=max_request_capacity,
                            max_token_capacity=max_token_capacity,
                            logging_fn=save_request)



import openai


class Model:
  """Abstraction over the OpenAI model, used to help maintain references to original functions."""

  def __init__(self, model_function):
    # Initialize instance variables
    self.model_function = model_function
    self.original_openai_chat = openai.ChatCompletion.create
    self.original_openai_completion = openai.Completion.create
    self.original_openai_embeddings = openai.Embedding.create

  def get_model_function(self):
    # get the model function
    return self.model_function

  def get_original_chat(self):
    return self.original_openai_chat

  def get_original_completion(self):
    return self.original_openai_completion

  def get_original_embeddings(self):
    return self.original_openai_embeddings


import time
from queue import Queue
import threading
import tiktoken


class CustomQueue(Queue):

  def __init__(self, max_token_capacity, max_request_capacity, verbose, *args,
               **kwargs):
    super().__init__(*args, **kwargs)
    self.verbose = verbose
    self.available_token_capacity = max_token_capacity
    self.available_request_capacity = max_request_capacity
    self.max_request_capacity = max_request_capacity
    self.max_token_capacity = max_token_capacity
    self.encoding = tiktoken.encoding_for_model("gpt-3.5-turbo")
    self.curr_time_block = int(time.time() // 60)
    self.lock = threading.Lock()

  def print_verbose(self, *args):
    if self.verbose:
      print(*args)

  def get(self, block=True, timeout=None):
    with self.lock:
      if self.is_empty():
        self.print_verbose("inside is empty")
        return super().get(block=block, timeout=timeout)
      # return the list of valid queries
      curr_time = time.time()
      # Convert it to minutes
      curr_time_minutes = int(curr_time // 60)
      self.print_verbose("curr_time_minutes: ", curr_time_minutes)
      self.print_verbose("self.curr_time_block: ", self.curr_time_block)
      #check how many requests have already been processed for that minute
      if self.curr_time_block == curr_time_minutes:
        self.print_verbose("in current time block")
        available_token_capacity = self.available_token_capacity
        available_request_capacity = self.available_request_capacity
        filtered_capacity = 0
        for args in list(self.queue)[:int(
            available_request_capacity
        )]:  # we can make a max number of requests per minute, check how much is left and iterate through the upperbound
          function, argument, task_id = args
          token_size = self.get_num_tokens(
            argument
          )  # assume you're getting the input prompt and the max output
          tmp_token_capacity = available_token_capacity - token_size
          if tmp_token_capacity >= 0:
            available_token_capacity = tmp_token_capacity
            filtered_capacity += 1
          else:
            break
        self.print_verbose("here's the filtered capacity: ", filtered_capacity)
        self.print_verbose("here's the previous available token capacity: ",
                           available_token_capacity)
        self.available_token_capacity = available_token_capacity
        self.available_request_capacity -= filtered_capacity
      else:
        available_token_capacity = self.max_token_capacity
        available_request_capacity = self.max_request_capacity
        filtered_capacity = 0
        #self.print_verbose(self.queue[:available_request_capacity])
        #self.print_verbose(type(self.queue[:available_request_capacity]))
        for args in list(self.queue)[:int(
            available_request_capacity
        )]:  # iterate through all available capacity until all chunks are done
          function, argument, task_id = args
          token_size = self.get_num_tokens(argument)
          self.print_verbose("token_size: ", token_size)
          tmp_token_capacity = available_token_capacity - token_size
          self.print_verbose("tmp_token_capacity: ", tmp_token_capacity)
          if tmp_token_capacity >= 0:
            available_token_capacity = tmp_token_capacity
            filtered_capacity += 1
          else:
            break
        self.print_verbose("here's the filtered capacity: ", filtered_capacity)
        self.print_verbose("here's the previous available token capacity: ",
                           available_token_capacity)
        self.available_token_capacity = available_token_capacity
        self.curr_time_block = curr_time_minutes
        self.available_request_capacity -= filtered_capacity
      # Create a new list with the required elements, using slicing
      new_list = list(self.queue)[:filtered_capacity]

      # Update the original list by removing the first `filtered_capacity` elements
      for _ in range(filtered_capacity):
        self.queue.popleft()
      self.print_verbose(
        "returning new list. Here's the remaining capacity for this minute: ",
        self.available_token_capacity)
      # Return the new list
      return new_list

  def get_num_tokens(self, chunk):
    return len(self.encoding.encode(chunk))

  def is_empty(self):
    return self.qsize() == 0


from termcolor import colored
import uuid
import time
import traceback


## Dev Imports 
# from IndividualRequest import IndividualRequest
# from RateLimitHandler_helpers.APICallHandler import APICallHandler

## Prod Imports
from reliablegpt.IndividualRequest import IndividualRequest
from reliablegpt.APICallHandler import APICallHandler


class RateLimitHandler:
  """A brief description of the class."""

  def __init__(
      self,
      model=None,
      fallback_strategy=[
        'gpt-3.5-turbo', 'text-davinci-003', 'gpt-4', 'text-davinci-002'
      ],
      graceful_string="Sorry, the OpenAI API is currently down",
      user_email="",
      user_token="",
      send_notification=False,
      max_token_capacity=40000,  #using default openai gpt-4 token capacity
      max_request_capacity=200,  #using default openai gpt-4 request capacity
      logging_fn=None,
      verbose=False):
    # Initialize instance variables
    super().__init__()
    self.model = model
    self.process_func = model.get_model_function()
    self.verbose = verbose
    self.user_email = user_email
    self.logging_fn = logging_fn
    self.graceful_string = graceful_string
    self.api_handler = APICallHandler(max_token_capacity,
                                      max_request_capacity,
                                      verbose=verbose)
    self.individual_request_handler = IndividualRequest(
      model,
      fallback_strategy=fallback_strategy,
      graceful_string=graceful_string,
      user_email=user_email,
      user_token=user_token,
      logging_fn=self.logging_fn,
      send_notification=False)

  def __call__(self, *args, **kwargs):
    try:
      result = self.get_request(kwargs)
      return result
    except:
      self.print_verbose(f"Hits an error: {traceback.format_exc()}")

  def print_verbose(self, print_statement):
    if self.verbose:
      print(colored("RateLimitHandler Request: " + print_statement, "yellow"))

  def get_batch_result_helper(self, text):
    task_id = uuid.uuid4().int
    self.print_verbose("task_id: ", task_id)
    self.print_verbose(
      f"function description: {self.individual_request_handler}")
    self.api_handler.add_task(process_func=self.individual_request_handler,
                              input=text,
                              task_id=task_id)

    start_time = time.time()

    while time.time() - start_time < self.set_timeout:
      result = self.api_handler.get_result(task_id)
    if result:
      return result

    return None

  def get_request(self, kwargs, set_timeout=1200):
    try:
      task_id = uuid.uuid4().int
      self.print_verbose(f"task_id: {task_id}")
      self.print_verbose(
        f"function description: {self.individual_request_handler}")
      self.api_handler.add_task(process_func=self.individual_request_handler,
                                input=kwargs,
                                task_id=task_id)

      start_time = time.time()
      while time.time() - start_time < set_timeout:
        results = self.api_handler.get_result(task_id=task_id)
        if results:
          return results

      result = self.graceful_string
      self.logging_fn(user_email=self.user_email,
                      graceful_string=self.graceful_string,
                      posthog_event='reliableGPT.recovered_request_exception',
                      result=result,
                      posthog_metadata={
                        'error': "Request timed out",
                        'recovered_response': result
                      },
                      errors=["Request timed out"])
      return result
    except:
      result = self.graceful_string
      self.logging_fn(user_email=self.user_email,
                      graceful_string=self.graceful_string,
                      posthog_event='reliableGPT.recovered_request_exception',
                      result=result,
                      posthog_metadata={
                        'error': traceback.format_exc(),
                        'recovered_response': result
                      },
                      errors=[traceback.format_exc()])
      return result

  def get_results(self):
    return self.api_handler.get_results()

  def get_result_threaded(self, test_question, results, set_timeout):
    self.print_verbose(f"Enters reliableGPT with question: {test_question}")
    start_time = time.time()
    result = self.get_batch_result_helper(test_question, set_timeout)
    results.append(result)
    end_time = time.time()
    self.print_verbose(f"Gets result from reliableGPT: {result}",
                       end_time - start_time)

  def batch_process(self, questions=[], set_timeout=1200):
    for question in questions:
      task_id = uuid.uuid4().int
      self.print_verbose("task_id: ", task_id)
      self.api_handler.add_task(process_func=self.process_func,
                                input=question,
                                task_id=task_id)
    start_time = time.time()
    while time.time() - start_time < set_timeout:
      results = self.api_handler.get_results()
      if len(results) >= len(questions):
        return results
    return self.api_handler.get_results()
